#
# httprequest.rb -- HTTPRequest Class
#
# Author: IPR -- Internet Programming with Ruby -- writers
# Copyright (c) 2000, 2001 TAKAHASHI Masayoshi, GOTOU Yuuzou
# Copyright (c) 2002 Internet Programming Ruby writers. All right reserved.
#
# $IPR: httprequest.rb,v 1.48 2002/09/08 19:47:58 gotoyuzo Exp $

require 'timeout'
require 'uri'

require 'webrick/httpstatus'
require 'webrick/httputils'
require 'webrick/cookie'

module WEBrick

  class HTTPRequest
    # Request line
    attr_reader :request_line
    attr_reader :request_method, :unparsed_uri, :http_version

    # Request-URI
    attr_reader :request_uri, :host, :path, :query_string, :query
    attr_accessor :script_name, :path_info
    alias :uri :path

    # Header and entity body
    attr_reader :raw_header, :header, :cookies
    attr_reader :body

    # Misc
    attr_reader :addr, :peeraddr
    attr_reader :attributes
    attr_reader :keep_alive
    attr_reader :request_time

    def initialize(config)
      @config = config
      @logger = config[:Logger]

      @request_line = @request_method =
        @unparsed_uri = @http_version = nil

      @request_uri = @host = @path = nil
      @script_name = @path_info = nil
      @query_string = nil
      @query = {}

      @raw_header = Array.new
      @header = Hash.new([].freeze)
      @cookies = []
      @body = nil

      @addr = @peeraddr = nil
      @request_time = nil
      @attributes = {}
      @keep_alive = false
    end

    def parse(socket=nil)
      @peeraddr = socket.respond_to?(:peeraddr) ? socket.peeraddr : []
      @addr = socket.respond_to?(:addr) ? socket.addr : []

      read_request_line(socket)
      read_header(socket)
      read_body(socket)

      return if @request_method == "CONNECT"

      begin
        @request_uri = make_absolute_request_uri()
        @path = HTTPUtils::unescape(@request_uri.path)
        @path = HTTPUtils::normalize_path(@path)
      rescue
        raise HTTPStatus::BadRequest, "bad URI `#@request_uri'."
      end

      @script_name = ""
      @path_info = @path.dup
      @query_string = @request_uri.query

      case @request_method
      when "GET", "HEAD"
        @query = HTTPUtils::parse_query(@query_string)
      else
        if self['content-type'] == "application/x-www-form-urlencoded"
          @query = HTTPUtils::parse_query(@body)
        end
      end

      @header['cookie'].each{|cookie|
        @cookies += Cookie::parse(cookie)
      }

      if /close/io =~ self["connection"]
        @keep_alive = false
      elsif /keep-alive/io =~ self["connection"]
        @keep_alive = true
      elsif @http_version < 1.1
        @keep_alive = false
      else
        @keep_alive = true
      end
    end

    def read_request_line(socket)
      @request_line = read_line(socket) if socket
      @request_time = Time.now
      raise HTTPStatus::EOFError unless @request_line
      if /^(\S+)\s+(\S+)(\s+HTTP\/(\d+\.\d+))?/o =~ @request_line
        @request_method = $1
        @unparsed_uri   = $2
        @http_version   = $4 ? $4.to_f : 0.9 
      else
        rl = @request_line.sub(/\x0d?\x0a\z/o, '')
        raise HTTPStatus::BadRequest, "bad Request-Line `#{rl}'."
      end
    end
    private :read_request_line

    def read_header(socket)
      if socket
        while line = read_line(socket)
          break if /\A(#{CRLF}|#{LF})\z/om =~ line
          @raw_header << line
        end
      end

      @header.clear
      field = nil
      @raw_header.each{|line|
        case line
        when /^([A-Za-z0-9_-]+):\s*(.*?)\s*\z/om
          field, value = $1, $2
          field.downcase!
          @header[field] = [] unless @header.has_key?(field)
          @header[field] << value
        when /^\s+(.*?)\s*\z/om
          value = $1
          unless field
            message =  "bad header '#{line.inspect}'."
            raise HTTPStatus::BadRequest, message
          end
          @header[field][-1] << " " << value
        else
          message =  "bad header '#{line.inspect}'."
          raise HTTPStatus::BadRequest, message
        end
      }

      @header.each{|key, values|
        values.each{|value|
          value.strip!
          value.gsub!(/\s+/, " ")
        }
      }
    end
    private :read_header

    def read_body(socket)
      return unless socket
      if tc = self['transfer-encoding']
        case tc
        when /chunked/io
          @body = read_chunked(socket)
        else
          raise HTTPStatus::NotImplemented, "Transfer-Encoding: #{tc}."
        end
      elsif self['content-length']
        size = self['content-length'].to_i
        @body = read_data(socket, size)
        if @body.nil? || @body.size != size
          raise HTTPStatus::BadRequest, "invalid body size."
        end
      elsif @request_method == "PUT" || @request_method == "POST"
        raise HTTPStatus::LengthRequired
      end
    end
    private :read_body

    def read_chunk_size(io)
      line = read_line(io)
      if /^([0-9a-fA-F]+)(?:;(\S+))?/ =~ line
        chunk_size = $1.hex
        chunk_ext = $2
        [ chunk_size, chunk_ext ]
      else
        raise HTTPStatus::BadRequest, "bad chunk `#{line}'."
      end
    end
    private :read_chunk_size

    def read_chunked(socket)
      body = ""
      chunk_size, = read_chunk_size(socket)
      while chunk_size > 0
        data = read_data(socket, chunk_size) # read chunk-data
        if data.nil? || data.size != chunk_size
          raise BadRequest, "bad chunk data size."
        end
        read_line(socket)              # skip CRLF
        body << data
        chunk_size, = read_chunk_size(socket)
      end
      read_header(socket)              # trailer + CRLF
      @header.delete("transfer-encoding")
      body
    end
    private :read_chunked

    def make_absolute_request_uri
      uri = URI::parse(@unparsed_uri)
      return uri if uri.absolute?

      if hostport = self["Host"]
        host, port = hostport.split(":", 2)
      elsif @addr.size > 0
        host, port = @addr[2], @addr[1]
      else
        host, port = @config[:ServerName], @config[:Port]
      end

      klass, scheme = URI::HTTP, "http"
      port = 80 unless port
      if @config[:SSLEnable]
        klass, scheme = URI::HTTPS, "https"
        port = 443 unless port
      end

      userinfo = uri.userinfo
      registry = opaque = fragment = nil

      klass.new(scheme, userinfo, host, port.to_s,
                registry, uri.path, opaque, uri.query, fragment)
    end

    def [](header_name)
      value = @header[header_name.downcase]
      value.empty? ? nil : value.join(", ")
    end

    def each
      @header.each{|k, v|
        value = @header[k]
        yield(k, value.empty? ? nil : value.join(", "))
      }
    end

    def keep_alive?
      @keep_alive
    end

    def to_s
      ret = @request_line.dup
      @raw_header.each{|line| ret << line }
      ret << CRLF
      ret << @body if @body
      ret
    end

    def meta_vars
      # This method provides the metavariables defined by the revision 3
      # of ``The WWW Common Gateway Interface Version 1.1''.
      # (http://Web.Golux.Com/coar/cgi/)

      meta = Hash.new

      if /^(Basic|Digest)\s+(.+)/o =~ self["Authorization"]
        auth_type = $1
        userpass = $2
        meta["AUTH_TYPE"]       = auth_type
        if auth_type == "Basic"
          remote_user, = decode64(userpass).split(":", 2)
          meta["REMOTE_USER"]   = remote_user
        end
      end

      cl = self["Content-Length"]
      ct = self["Content-Type"]
      meta["CONTENT_LENGTH"]    = cl if cl.to_i > 0
      meta["CONTENT_TYPE"]      = ct.dup if ct
      meta["GATEWAY_INTERFACE"] = "CGI/1.1"
      meta["PATH_INFO"]         = @path_info.dup
     #meta["PATH_TRANSLATED"]   = nil      # no plan to be provided
      meta["QUERY_STRING"]      = @query_string ? @query_string.dup : ""
      meta["REMOTE_ADDR"]       = @peeraddr[3]
      meta["REMOTE_HOST"]       = @peeraddr[2]
     #meta["REMOTE_IDENT"]      = nil      # no plan to be provided
      meta["REQUEST_METHOD"]    = @request_method.dup
      meta["REQUEST_URI"]       = @request_uri.to_s
      meta["SCRIPT_NAME"]       = @script_name.dup
      meta["SERVER_NAME"]       = @request_uri.host
      meta["SERVER_PORT"]       = @config[:Port].to_s
      meta["SERVER_PROTOCOL"]   = "HTTP/" + @config[:HTTPVersion].to_s
      meta["SERVER_SOFTWARE"]   = @config[:ServerSoftware].dup
      meta["HTTPS"]             = "on" if @config[:SSLEnable]

      self.each{|key, val|
        name = "HTTP_" + key
        name.gsub!(/-/o, "_")
        name.upcase!
        meta[name] = val
      }

      meta
    end

    def _read_data(io, method, arg)
      begin
        timeout(@config[:RequestTimeout]){
          return io.__send__(method, arg)
        }
      rescue Errno::ECONNRESET
        return nil
      rescue TimeoutError
        raise HTTPStatus::RequestTimeout
      end
    end

    def read_line(io)
      _read_data(io, :gets, LF)
    end

    def read_data(io, size)
      _read_data(io, :read, size)
    end

    private :_read_data, :read_line, :read_data
  end
end
