#
# httpresponse.rb -- HTTPResponse Class
#
# Author: IPR -- Internet Programming with Ruby -- writers
# Copyright (c) 2000, 2001 TAKAHASHI Masayoshi, GOTOU Yuuzou
# Copyright (c) 2002 Internet Programming Ruby writers. All right reserved.
#
# $IPR: httpresponse.rb,v 1.34 2002/09/08 19:47:58 gotoyuzo Exp $

require 'time'
require 'webrick/htmlutils'
require 'webrick/httputils'
require 'webrick/httpstatus'

module WEBrick
  class HTTPResponse
    BUFSIZE = 1024*64

    attr_reader :status, :header, :cookies
    attr_accessor :http_version
    attr_accessor :reason_phrase
    attr_accessor :body
    attr_accessor :keep_alive
    attr_accessor :request_uri
    attr_accessor :filename

    attr_reader :config, :sent_size

    def initialize(config)
      @config = config
      @header = Hash.new
      @status = HTTPStatus::RC_OK
      @reason_phrase = nil
      @http_version = @config[:HTTPVersion]
      @body = ''
      @keep_alive = true
      @cookies = []
      @request_uri = nil
      @chunked = false
      @filename = nil
      @sent_size = 0
    end

    def status=(status)
      @status = status
      @reason_phrase = HTTPStatus::reason_phrase(status)
    end

    def [](field)
      @header[field.downcase]
    end

    def []=(field, value)
      @header[field.downcase] = value.to_s
    end

    def each
      @header.each{|k, v|  yield(k, v) }
    end

    def send_response(socket)
      send_header(socket) && send_body(socket)
    end

    def send_header(socket)
      @reason_phrase ||= HTTPStatus::reason_phrase(@status)
      @header['server'] ||= @config[:ServerSoftware]
      @header['date'] ||= Time.now.httpdate
      if chunked?
        @header["transfer-encoding"] = "chunked"
        @header.delete('content-length')
      else
        case @body
        when IO  
          unless @header['content-length']
            @keep_alive = false
          end
        when nil
          @header['content-length'] = 0
        else             # maybe a String
          @header['content-length'] = @body.size
        end
      end
      @header['connection'] = @keep_alive ? "Keep-Alive" : "close"
      if location = @header['location']
        if @request_uri
          @header['location'] = @request_uri.merge(location)
        end
      end

      h = status_line()
      @header.each{|key, value|
        ktmp = key.gsub(/^\w|-\w/){ |s| s.upcase }
        h << "#{ktmp}: #{value}" << CRLF
      }
      @cookies.each{|cookie|
        h << "Set-Cookie: " << cookie.to_s << CRLF
      }
      h << CRLF
      _write_data(socket, h)
    end

    def status_line
      "HTTP/#@http_version #@status #@reason_phrase #{CRLF}"
    end

    def send_body(socket)
      case @body
      when IO then send_body_io(socket)
      else
        if chunked?
          if @body && @body.size > 0
            socket << @body.size << CRLF
            _write_data(socket, @body)
            socket << CRLF
          end
          socket << 0 << CRLF
          socket << CRLF
        else             # maybe a String
          if @body && @body.size > 0
            _write_data(socket, @body)
          end
        end
        @sent_size = @body ? @body.size : 0
      end
    end

    def send_body_io(socket)
      begin
        if chunked?
          while buf = @body.read(BUFSIZE)
            socket << format("%x", buf.size) << CRLF
            socket << buf << CRLF
            @sent_size += buf.size
          end
          socket << 0 << CRLF
          socket << CRLF
        else
          size = @header['content-length'].to_i
          _send_file(socket, @body, 0, size.to_i)
          @sent_size = size
        end
      ensure
        @body.close
      end
    end
    private :send_body_io

    def _send_file(output, input, offset, size)
      while offset > 0
        sz = BUFSIZE < offset ? BUFSIZE : offset
        buf = input.read(sz)
        offset -= buf.size
      end

      if size > 0
        while size > 0
          sz = BUFSIZE < size ? BUFSIZE : size
          buf = input.read(sz)
          size -= buf.size
          output << buf
        end
      else
        while buf = input.read(BUFSIZE)
          output << buf
        end
      end
    end
    private :_send_file

    def _write_data(socket, data)
      begin
        socket << data
        return true
      rescue Errno::EPIPE
        @status = HTTPStatus::RC_INTERNAL_SERVER_ERROR
        @reason_phrase = "EPIPE occured"
        @keep_alive = false
        return false
      end
    end
    private :_write_data

    def chunked?
      @chunked
    end

    def chunked=(val)
      @chunked = val ? true : false
    end

    def to_s
      ret = ""
      send_response(ret)
      ret
    end

    def set_redirect(status, url)
      @body = "<HTML><A HREF=\"#{url.to_s}\">#{url.to_s}</A>.</HTML>\n"
      @header['location'] = url.to_s
      raise status
    end

    def set_error(ex, backtrace=false)
      case ex
      when HTTPStatus::Status 
        @keep_alive = false if HTTPStatus::error?(ex.code)
        self.status = ex.code
      else 
        @keep_alive = false
        self.status = HTTPStatus::RC_INTERNAL_SERVER_ERROR
      end
      @header['content-type'] = "text/html"

      if respond_to?(:create_error_page)
        create_error_page()
        return
      end

      if @request_uri
        host, port = @request_uri.host, @request_uri.port
      else
        host, port = @config[:ServerName], @config[:Port]
      end

      @body = ''
      @body << <<-_end_of_html_
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN">
<HTML>
  <HEAD><TITLE>#{HTMLUtils::escape(@reason_phrase)}</TITLE></HEAD>
  <BODY>
    <H1>#{HTMLUtils::escape(@reason_phrase)}</H1>
    #{HTMLUtils::escape(ex.message)}
    <HR>
      _end_of_html_

      if backtrace && $DEBUG
        @body << "backtrace of `#{HTMLUtils::escape(ex.type.to_s)}' "
        @body << "#{HTMLUtils::escape(ex.message)}"
        @body << "<PRE>"
        ex.backtrace.each{|line| @body << "\t#{line}\n"}
        @body << "</PRE><HR>"
      end

      @body << <<-_end_of_html_
    <ADDRESS>
     #{HTMLUtils::escape(@config[:ServerSoftware])} at
     #{host}:#{port}
    </ADDRESS>
  </BODY>
</HTML>
      _end_of_html_
    end

    def keep_alive?
      @keep_alive
    end

  end
end
